namespace Tuned {
  namespace AudioTools {
    namespace Waveform {
      namespace ExitStatus {
        public const int OK = 0;
        public const int INVALID_COMMANDLINE_PARAMETERS = 1;
        public const int DECODE_ERROR = 2;
        public const int WRITE_ERROR = 3;
        public const int NO_CHANNELS_ERROR = 4;
        public const int TOO_MANY_CHANNELS_ERROR = 5;
        public const int ONLY_PRINT_VERSION = 255;
      }
      
      public class App {
        static const string VERSION = "1.0.2";
        
        static bool     version = false;
        
        static bool     audio_input_stdin = false;
        static string   audio_input_file;
        static string   audio_calculation_method;

        static string   audio_calculation_method_level;
        static double   audio_volume_floor_level;
        
        static Gee.ArrayList<Gee.ArrayList<uint16>> audio_data;
        
        static bool     waveform_output_stdout = false;
        static bool     waveform_output_stderr = false;
        static string   waveform_output_file;
        static string   waveform_output_format;

        static int      waveform_pixels_per_second = 100;
        static int      waveform_range_channel;
        static int      waveform_range_total;
        static string   waveform_background_color;
        static double   waveform_background_color_red;
        static double   waveform_background_color_green;
        static double   waveform_background_color_blue;

        static string   waveform_foreground_color;
        static double   waveform_foreground_color_red;
        static double   waveform_foreground_color_green;
        static double   waveform_foreground_color_blue;
        
        static uint64   interval;
        
        static MainLoop mainloop;

	      static dynamic Gst.Pipeline pipeline;
	      static dynamic Gst.Element  source;
	      static dynamic Gst.Element  decoder;
	      static dynamic Gst.Element  level;
	      static dynamic Gst.Element  sink;
	      
	      
	      // Temporary variables
	      static int      audio_rate;
	      static int      audio_channels;

	      static uint16   waveform_height_per_channel;
	      
      
        const OptionEntry[] commandline_app_entries = 
        {
          { "audio-input-stdin",          0, 0,  OptionArg.NONE,   out audio_input_stdin,       "Read the audio data from a standard input", null },
          { "audio-input-file",           0, 0,  OptionArg.STRING, out audio_input_file,        "Read the audio data from a file specified by PATH", "PATH" },
          { "audio-calculation-method",   0, 0,  OptionArg.STRING, out audio_calculation_method,"Audio calculation method, one of \"volume-peak\" (default), \"volume-rms\"", "METHOD" },
          { "audio-volume-floor-level",   0, 0,  OptionArg.DOUBLE, out audio_volume_floor_level,"Level of silence while audio calculation method is set to \"volume-peak\" or \"volume-rms\"  (defaults to -60 dB)", "LEVEL" },

          { "waveform-output-stdout",     0, 0,  OptionArg.NONE,   out waveform_output_stdout,     "Write final image containing the waveform to the standard output", null },
          { "waveform-output-stderr",     0, 0,  OptionArg.NONE,   out waveform_output_stderr,     "Write final image containing the waveform to the standard error output", null },
          { "waveform-output-file",       0, 0,  OptionArg.STRING, out waveform_output_file,       "Write final image containing the waveform to the file specified by PATH", "PATH" },
          { "waveform-output-format",     0, 0,  OptionArg.STRING, out waveform_output_format,     "Write final image data in a one of the following formats: \"png\" (default), \"xml\",\"json\", \"plain\"", "FORMAT" },

          { "waveform-pixels-per-second", 0, 0, OptionArg.INT,     out waveform_pixels_per_second, "How many pixels should represent one second of the audio data (in pixels, defaults to 100)", "INT" },
          { "waveform-range-channel",     0 ,0, OptionArg.INT,     out waveform_range_channel, "Set range of the values calculated by the algorithm to 0..INT per channel. In case of PNG output it means that output image height = amount of channels * INT", "INT" },
          { "waveform-range-total",       0 ,0, OptionArg.INT,     out waveform_range_total, "Set range of the values calculated by the algorithm to 0..INT in total. Equal to the output image height in case of PNG output, regardless of amount of channels.", "INT" },

          { "waveform-background-color",  0, 0, OptionArg.STRING,  out waveform_background_color,  "Image background color, in RGB hex notation (defaults to \"#ffffff\")", "RGB" },
          { "waveform-foreground-color",  0, 0, OptionArg.STRING,  out waveform_foreground_color,  "Image foreground color, in RGB hex notation (defaults to \"#666666\")", "RGB" },

          { "version",                    0, 0, OptionArg.NONE,    out version,                 "Print version and exit", null },
          
          { null }
        };

      
        static int main(string[] args) {
          try {

            var opt = new OptionContext(" - Draw a waveform of the audio file");
            opt.set_help_enabled(true);
            opt.add_main_entries(commandline_app_entries, null);

            var gst_opt_group = Gst.init_get_option_group();
            opt.add_group((owned)gst_opt_group);
          
            opt.parse(ref args);
            
          } catch (GLib.Error e) {
            stderr.printf("Error: %s\n", e.message);
            stderr.printf("Run '%s --help' to see a full list of available options\n", args[0]);
            Process.exit(ExitStatus.INVALID_COMMANDLINE_PARAMETERS);
          }
          
          if(version == true) {
            stdout.printf("%s\n", VERSION);
            Process.exit(ExitStatus.ONLY_PRINT_VERSION);
          }

          if((audio_input_file == null || audio_input_file == "") && audio_input_stdin == false) {
            stderr.printf("Error: You must set desired input for the audio data.\n");
            stderr.printf("Run '%s --help' to see a full list of available options\n", args[0]);
            Process.exit(ExitStatus.INVALID_COMMANDLINE_PARAMETERS);
          }


          if((waveform_output_file == null || waveform_output_file == "") && waveform_output_stdout == false && waveform_output_stderr == false) {
            stderr.printf("Error: You must set desired output for the image containing the waveform.\n");
            stderr.printf("Run '%s --help' to see a full list of available options\n", args[0]);
            Process.exit(ExitStatus.INVALID_COMMANDLINE_PARAMETERS);
          }

          if(waveform_output_format == null || waveform_output_format == "") {
            waveform_output_format = "png";
          
          } else if (waveform_output_format != "png" &&  waveform_output_format != "json" && waveform_output_format != "xml" && waveform_output_format != "plain") {
            stderr.printf("Error: Image output format \"%s\" is unknown.\n", waveform_output_format);
            stderr.printf("Run '%s --help' to see a full list of available options\n", args[0]);
            Process.exit(ExitStatus.INVALID_COMMANDLINE_PARAMETERS);
          }

          if(audio_calculation_method == null || audio_calculation_method == "" || audio_calculation_method == "volume-peak") {
            audio_calculation_method_level = "peak";

          } else if(audio_calculation_method == "volume-rms") {
            audio_calculation_method_level = "rms";
            
          } else {
            stderr.printf("Error: Audio calculation method \"%s\" is unknown.\n", audio_calculation_method);
            stderr.printf("Run '%s --help' to see a full list of available options\n", args[0]);
            Process.exit(ExitStatus.INVALID_COMMANDLINE_PARAMETERS);
          
          }
          
          if(waveform_pixels_per_second <= 0) {
            stderr.printf("Error: Pixels per second has to be positive value.\n");
            stderr.printf("Run '%s --help' to see a full list of available options\n", args[0]);
            Process.exit(ExitStatus.INVALID_COMMANDLINE_PARAMETERS);
          }


          if(audio_volume_floor_level > 0) {
            stderr.printf("Error: Audio volume floor level has to be negative value.\n");
            stderr.printf("Run '%s --help' to see a full list of available options\n", args[0]);
            Process.exit(ExitStatus.INVALID_COMMANDLINE_PARAMETERS);

          } else if(audio_volume_floor_level == 0)  {
            audio_volume_floor_level = -60.0;
          }

          if(waveform_range_total == 0 && waveform_range_channel == 0) {
            stderr.printf("Error: You must set desired channel or total calculation range, please use --waveform-range-total or --waveform-range-channel.\n");
            stderr.printf("Run '%s --help' to see a full list of available options\n", args[0]);
            Process.exit(ExitStatus.INVALID_COMMANDLINE_PARAMETERS);
          }
          
          if(waveform_range_total > 0 && waveform_range_channel > 0) {
            stderr.printf("Error: You cannot set both --waveform-range-total and --waveform-range-channel.\n");
            stderr.printf("Run '%s --help' to see a full list of available options\n", args[0]);
            Process.exit(ExitStatus.INVALID_COMMANDLINE_PARAMETERS);
          }
          
          if(waveform_range_channel == 0 && waveform_range_total < 0) {
            stderr.printf("Error: Waveform total range must be a positive value.\n");
            stderr.printf("Run '%s --help' to see a full list of available options\n", args[0]);
            Process.exit(ExitStatus.INVALID_COMMANDLINE_PARAMETERS);
          }


          if(waveform_range_total == 0 && waveform_range_channel < 0) {
            stderr.printf("Error: Waveform channel range must be a positive value.\n");
            stderr.printf("Run '%s --help' to see a full list of available options\n", args[0]);
            Process.exit(ExitStatus.INVALID_COMMANDLINE_PARAMETERS);
          }
          
          if(waveform_range_total > uint16.MAX) {
            stderr.printf("Error: Image height cannot be bigger than " + uint16.MAX.to_string() + ".\n");
            stderr.printf("Run '%s --help' to see a full list of available options\n", args[0]);
            Process.exit(ExitStatus.INVALID_COMMANDLINE_PARAMETERS);
          }
          
          if(waveform_background_color == null || waveform_background_color == "") {
            waveform_background_color_red   = 1.0;
            waveform_background_color_green = 1.0;
            waveform_background_color_blue  = 1.0;

          } else if (/^\#[0-9a-fA-F]{6}$/.match(waveform_background_color)) {
            waveform_background_color_red   = (waveform_background_color[1].xdigit_value() * 16 + waveform_background_color[2].xdigit_value()) / 255.0;
            waveform_background_color_green = (waveform_background_color[3].xdigit_value() * 16 + waveform_background_color[4].xdigit_value()) / 255.0;
            waveform_background_color_blue  = (waveform_background_color[5].xdigit_value() * 16 + waveform_background_color[6].xdigit_value()) / 255.0;

          } else {
            stderr.printf("Error: Value \"%s\" specified as image's background color is not valid hexadecimal notation of a RGB color.\n", waveform_background_color);
            stderr.printf("Run '%s --help' to see a full list of available options\n", args[0]);
            Process.exit(ExitStatus.INVALID_COMMANDLINE_PARAMETERS);
          }

          if(waveform_foreground_color == null || waveform_foreground_color == "") {
            waveform_foreground_color_red   = 0.4;
            waveform_foreground_color_green = 0.4;
            waveform_foreground_color_blue  = 0.4;

          } else if (/^\#[0-9a-fA-F]{6}$/.match(waveform_foreground_color)) {
            waveform_foreground_color_red   = (waveform_foreground_color[1].xdigit_value() * 16 + waveform_foreground_color[2].xdigit_value()) / 255.0;
            waveform_foreground_color_green = (waveform_foreground_color[3].xdigit_value() * 16 + waveform_foreground_color[4].xdigit_value()) / 255.0;
            waveform_foreground_color_blue  = (waveform_foreground_color[5].xdigit_value() * 16 + waveform_foreground_color[6].xdigit_value()) / 255.0;

          } else {
            stderr.printf("Error: Value \"%s\" specified as image's foreground color is not valid hexadecimal notation of a RGB color.\n", waveform_foreground_color);
            stderr.printf("Run '%s --help' to see a full list of available options\n", args[0]);
            Process.exit(ExitStatus.INVALID_COMMANDLINE_PARAMETERS);
          }


		      Gst.init(ref args);
		      
		      audio_data = new Gee.ArrayList<Gee.ArrayList>();
		      
		      // Pipeline
		      pipeline = new Gst.Pipeline("pipeline");
 		      decoder  = Gst.ElementFactory.make("decodebin2", "decoder");
 		      level    = Gst.ElementFactory.make("level",      "levelmeter");
 		      sink     = Gst.ElementFactory.make("fakesink",   "sink");

          interval = calculate_interval();

 		      level.interval = interval;
 		      level.message = true;
 		      
		      
		      if(audio_input_stdin == true) {
		        source = Gst.ElementFactory.make("fdsrc", "source");
		        source.fd = stdin.fileno();

		      } else if(audio_input_file != null && audio_input_file != "") {
		        source = Gst.ElementFactory.make("filesrc", "source");
		        source.location = audio_input_file;

		      } else {
		        assert_not_reached();
		      }
		      
		      
		      
		      pipeline.add(source);
		      pipeline.add(decoder);
		      pipeline.add(level);
		      pipeline.add(sink);
		      
		      source.link(decoder);
		      level.link(sink);

		      decoder.pad_added.connect((element, pad) => {
//            debug("Linking receiver part of the pipeline with playback part of the pipeline");
            
            Gst.Pad output_pad = level.get_static_pad("sink");
            pad.link(output_pad);
          });	
		      

          Gst.Bus bus = pipeline.get_bus();
          bus.add_watch(bus_cb);		      

          Timeout.add(1, () => {
//            debug("Starting pipeline");
            pipeline.set_state(Gst.State.PLAYING);
            return false;
          });

          mainloop = new MainLoop();
          mainloop.run();
        
          return ExitStatus.OK;
        }
        
        static uint64 calculate_interval() {
          return (uint64)(1 * Gst.SECOND / waveform_pixels_per_second);
        }
        
        static bool bus_cb(Gst.Bus bus, Gst.Message message) {
          switch (message.type) {
            case Gst.MessageType.ERROR:
              GLib.Error err;
              string debug;
              message.parse_error (out err, out debug);
              stderr.printf("Error: Decoding error from element %s: %s\n", message.src.get_name(), err.message);
              Process.exit(ExitStatus.DECODE_ERROR);
              break;


            case Gst.MessageType.EOS:
              switch(waveform_output_format) {
                case "json":
                  write_waveform_json();
                  break;
                  
                case "xml":
                  write_waveform_xml();
                  break;

                case "png":
                  write_waveform_png();
                  break;

                case "plain":
                  write_waveform_plain();
                  break;
                
              }

              mainloop.quit();
              break;
              
            case Gst.MessageType.ELEMENT:
              if(message.src == level) {
                
                var str = message.get_structure();
                var channels = str.get_value(audio_calculation_method_level);

                Gst.ClockTime running_time;
                
                str.get_clock_time("running-time", out running_time);
                
                running_time = running_time / interval;

//                debug("Interval: " + interval.to_string() + ", RunningTime: " + running_time.to_string() + ", Size: " + audio_data[0].size.to_string());

                for(var channel = 0; channel < audio_channels; channel++) {
                  double channel_level_db = channels.list_get_value(channel).get_double();
                  
                  // Take care of skipped samples
                  if(audio_data[channel].size != 0) {
                    while(audio_data[channel].size < running_time) {
                      audio_data[channel].add(audio_data[channel][audio_data[channel].size-1]);
                    }
                  }
                  
                  // TODO check for inifinity and NaN
                  if(channel_level_db < audio_volume_floor_level) {
                    audio_data[channel].add(0);
//                    debug("channel %u = silence", channel);

                  } else {
                    uint16 channel_level_int = (uint16)(waveform_height_per_channel * ((audio_volume_floor_level - channel_level_db) / audio_volume_floor_level));
//                    debug("channel %u = %.2f dB => " + channel_level_int.to_string(), channel, channel_level_db);
                    
                    audio_data[channel].add(channel_level_int);
                  }

                  
                }
                
              }
              
              break;
              


                
            case Gst.MessageType.STATE_CHANGED:
              Gst.State oldstate;
              Gst.State newstate;
              Gst.State pending;
              message.parse_state_changed(out oldstate, out newstate, out pending);

//              debug("Element %s state changed: %s -> %s : %s", message.src.get_name(), oldstate.to_string(), newstate.to_string(), pending.to_string());

              if(message.src == pipeline && newstate == Gst.State.PAUSED) {
//                debug("Trying to read metadata");

                Gst.Pad pad = sink.get_static_pad("sink");

                Gst.Caps caps = pad.get_negotiated_caps();
                if(caps != null) {
                  Gst.Structure str = caps.get_structure(0);
                  
                  str.get_int("rate", out audio_rate);
                  str.get_int("channels", out audio_channels);
                  
                  if(audio_channels == 0) {
                    stderr.printf("Error: No channels found in the decoded audio.\n");
                    Process.exit(ExitStatus.NO_CHANNELS_ERROR);
                  }
                
//                  debug("Rate: %d Hz", audio_rate);
//                  debug("Channels: %d", audio_channels);
                  
                  for(var i = 0; i < audio_channels; i++) {
                    var channel_data = new Gee.ArrayList<uint16>();
                    
                    audio_data.add(channel_data);
                  }
                  
                  // Just save this variable to avoid doing that in every iteration of the level message callback                  
                  if(is_waveform_range_specified_as_total()) {
                    waveform_height_per_channel = (uint16)(waveform_range_total / audio_channels);
                    // FIXME check if it is a right way to convert from int to uint16
                  } else {
                    waveform_height_per_channel = (uint16)waveform_range_channel;
                  }
                  
                } else {
                  stderr.printf("Error: Cannot read audio metadata.\n");
                  Process.exit(ExitStatus.DECODE_ERROR);
                }
              
              }

              
              break;

                
            default:
              break;
          }


          return true;
        }         
        
        
        static bool is_waveform_range_specified_as_total() {
          return waveform_range_total != 0 && waveform_range_channel == 0;
        }

        
        static int waveform_height() {
          return is_waveform_range_specified_as_total() ? waveform_range_total : waveform_range_channel * audio_channels   ;     
        }
        
        
        static void write_waveform_json() {
          bool first_channel = true;
          var output = new StringBuilder();
          output.append("{\n \"settings\": { \"pixels-per-second\" : ");
          output.append(waveform_pixels_per_second.to_string());
          output.append(" },\n \"channels\" : [");

          
          for(var channel = 0; channel < audio_channels; channel ++) {
            
            var channel_output = new StringBuilder();
            channel_output.append("\n  { \"index\" : ");
            channel_output.append(channel.to_string());
            channel_output.append(", \"range-min\" : 0, \"range-max\" : ");
            channel_output.append(waveform_height_per_channel.to_string());
            channel_output.append(", \"data\" : \n    [");
            
            bool first_pixel = true;

            foreach(var pixel in audio_data[channel]) {
              if(first_pixel == false)
                channel_output.append(",");

              channel_output.append(pixel.to_string());
              
              first_pixel = false;
            }
            channel_output.append("] }");
            
            if(first_channel == false)
              output.append(",\n");
              
            first_channel = false;

            output.append(channel_output.str);

          }
          
          output.append("\n ]\n}\n");
          
          write_waveform_text(output.str);
        }     
        
        static void write_waveform_xml() {
          var output = new StringBuilder();
          output.append("<?xml version=\"1.0\">\n<waveform>\n <settings>\n  <pixels-per-second>");
          output.append(waveform_pixels_per_second.to_string());
          output.append("</pixels-per-second>\n </settings>\n <channels>");
          
          for(var channel = 0; channel < audio_channels; channel ++) {
            
            var channel_output = new StringBuilder();
            channel_output.append("\n  <channel index=\"");
            channel_output.append(channel.to_string());
            channel_output.append("\" range-min=\"0\" range-max=\"");
            channel_output.append(waveform_height_per_channel.to_string());
            channel_output.append("\">\n");
            

            foreach(var pixel in audio_data[channel]) {
              channel_output.append("   <p>");
              channel_output.append(pixel.to_string());
              channel_output.append("</p>\n");
            }
            channel_output.append("  </channel>\n");

            output.append(channel_output.str);

          }
          
          output.append(" </channels>\n</waveform>\n");
          
          write_waveform_text(output.str);        
        }


        static void write_waveform_plain() {
          var output = new StringBuilder();
          
          for(var i = 0; i < audio_data[0].size; i++) {
            for(var channel = 0; channel < audio_channels; channel ++) {
              if(channel > 0)
                output.append("\t");

              output.append(audio_data[channel][i].to_string());
            }
            output.append("\n");
          }
          

          write_waveform_text(output.str);        
        }

        
        static void write_waveform_text(string contents) {
          if(waveform_output_stdout == true) {
            stdout.printf(contents);
          } else if(waveform_output_stderr == true) {
            stderr.printf(contents);
          
          } else {
            try {

              var file = File.new_for_path(waveform_output_file);
              if (file.query_exists()) {
                file.delete();
              }
              
              file.create(FileCreateFlags.REPLACE_DESTINATION, null);
              
              file.replace_contents(contents, contents.length, null, false, FileCreateFlags.REPLACE_DESTINATION, null, null);

            } catch(Error e) {
              // FIXME TODO does not understand ~ in the path
              stderr.printf("Error: Cannot write to file \"%s\": %s\n", waveform_output_file, e.message);
              Process.exit(ExitStatus.WRITE_ERROR);
            }
          }
        }



        static void write_waveform_png() {
          var height = waveform_height();
          var surface = new Cairo.ImageSurface(Cairo.Format.RGB24, audio_data[0].size, height);
          var context = new Cairo.Context(surface);
          
          // Background
          context.set_source_rgb(waveform_background_color_red, waveform_background_color_green, waveform_background_color_blue);
          context.move_to(0, 0);
          context.line_to(audio_data[0].size, 0);
          context.line_to(audio_data[0].size, height);
          context.line_to(0, height);
          context.close_path();
          context.fill();

          
          // Channel waveforms
          context.set_source_rgb(waveform_foreground_color_red, waveform_foreground_color_green, waveform_foreground_color_blue);
          context.set_line_width(1.0);

          for(var channel = 0; channel < audio_channels; channel ++) {
            var channel_y = (channel + 1) * waveform_height_per_channel;
            
            var x = 0;
            foreach(var pixel in audio_data[channel]) {
              context.move_to(x, channel_y);
              context.rel_line_to(0, -1 * pixel);
              context.stroke();

              x++;
            }
          }
          
          if(waveform_output_stdout) {
            surface.write_to_png_stream(write_waveform_png_stdout);

          } else if(waveform_output_stderr) {
            surface.write_to_png_stream(write_waveform_png_stderr);
            
          } else {
            surface.write_to_png(waveform_output_file);          
          }
        }

        static Cairo.Status write_waveform_png_stderr(uchar[] data) {
          foreach(var c in data) {
            stderr.printf("%c", (uchar)c);
          }
          
          return Cairo.Status.SUCCESS;
        }
        
        static Cairo.Status write_waveform_png_stdout(uchar[] data) {
          foreach(var c in data) {
            stdout.printf("%c", (uchar)c);
          }

          return Cairo.Status.SUCCESS;
        }
      }
    }
  }
}
